import numpy as np
import pandas as pd
import os
from sklearn.model_selection import train_test_split
from pykeen.triples import TriplesFactory
from pykeen.pipeline import pipeline
import pyiwn
from itertools import combinations
from pykeen.models import predict
import pickle5 as pickle
import torch
import networkx as nx

beng_iwn = pyiwn.IndoWordNet(lang = pyiwn.Language.BENGALI)
all_syns = beng_iwn.all_synsets()

edges = pd.read_csv("../data/BengIwnData/edges.csv", header = None, delimiter = '\t', names = ['ent1','ent2', 'rel'])

X = pd.DataFrame()
X['ent1'] = edges['ent1']
X['rel'] = edges['rel']
X['ent2'] = edges['ent2']
print("The dataframe X: \n", X.head())
X = X.astype(str)

print("Value counts for relation:")
print(X['rel'].value_counts())

print("Changing the edge structure from (synset -> synset) to (word -> word)")

# converting synsets list into a hashmap
syn_dict = dict()
for val in all_syns:
  syn_dict[val.synset_id()] = val
  
def generate_triples(tuple_list, rel_id):
  triple_list = []
  for tup in tuple_list:
    triple = (tup[0], rel_id, tup[1])
    triple_list.append(triple)
  return triple_list

def generate_allcombs(list1, list2):
  res = []
  for val1 in list1:
    for val2 in list2:
      tup = (val1, val2)
      res.append(tup)
  return res
  
  
word_edges = set()
visited_syns = set()
X = X.astype(int)
for i, row in X.iterrows():
  syn1_id = row['ent1']
  syn1 = syn_dict[syn1_id]
  word_list1 = syn1.lemma_names()
  rel = row['rel']
  syn2_id = row['ent2']
  syn2 = syn_dict[syn2_id]
  word_list2 = syn2.lemma_names()

  if(syn1_id not in visited_syns):
    # generating tuples for synonymy relation - id = 31
    synonymy_combs1 = combinations(word_list1, 2)
    synonymy_triples1 = generate_triples(synonymy_combs1, '31') # add to set
    for i, val in enumerate(synonymy_triples1):
      word_edges.add(val)
    visited_syns.add(syn1_id)

  if(syn2_id not in visited_syns):
    synonymy_combs2 = combinations(word_list2, 2)
    synonymy_triples2 = generate_triples(synonymy_combs2, '31') # add to set
    for i, val in enumerate(synonymy_triples2):
      word_edges.add(val)
    visited_syns.add(syn2_id)

  # enumerating the same relation for all words in both synset lists
  word_combinations = generate_allcombs(word_list1, word_list2)
  word_triples = generate_triples(word_combinations, str(rel)) # add to set

  for i, val in enumerate(word_triples):
    word_edges.add(val)

print(f"Number of edges with words as nodes = {len(word_edges)}")
word_edgelist = list(word_edges)  
triples = np.asarray(word_edgelist)

print("triples' shape: ", triples.shape)

# entity dictionary
words = beng_iwn.all_words()
ent_to_id = dict()
i = 0
for word in words:
  ent_to_id[word] = i
  i+=1

# relation dictionary
rel_to_id = dict()
for i in range(1,32):
  rel_to_id[str(i)] = i


tf = TriplesFactory.from_labeled_triples(triples)
print(tf)

train, test, val = tf.split([0.8, 0.1, 0.1])

# this script uses the predict_triples_df function 

from pykeen.hpo import hpo_pipeline

#pipeline_result_distmult = pipeline(
#    training = train,
#    testing = test,
#    validation = val,
#    model = 'DistMult',
#    model_kwargs=dict(embedding_dim=100),
#    evaluator = 'rankbased',
#    loss = "marginranking",
#    loss_kwargs = dict(margin=1, margin_activation='relu'),
#    negative_sampler_kwargs = dict(num_negs_per_epoch = 50),
#    optimizer ='adagrad',
#    optimizer_kwargs=dict(lr = 0.1),
#    regularizer="lp",
#    regularizer_kwargs=dict(weight=0.001),
#    training_kwargs = dict(num_epochs =10, batch_size=1932),
#    training_loop='slcwa',
#    stopper='early'
#)

#pipeline_result_distmult.save_to_directory("./distmult/best_model_ampli")

#hole
#pipeline_result_hole = pipeline(
#    training = train,
#    testing = test,
#    validation = val,
#    model = 'HolE',
#    model_kwargs=dict(embedding_dim=100),
#    evaluator = 'rankbased',
#    loss = "marginranking",
#    loss_kwargs = dict(margin=1, margin_activation='softplus'),
#    negative_sampler = 'basic',
#    optimizer ='adagrad',
#    optimizer_kwargs=dict(lr = 0.1),
#    training_loop='slcwa',
#    training_kwargs = dict(num_epochs =10, batch_size=3032),
#    stopper='early'
#)

# reading subgraph
subgraph_file = open("./DT_sub_iwn.gpickle", "rb")
subgraph = pickle.load(subgraph_file)

def generate_all_rel(u,v):
    result = []
    for rel_int in range(1,32):
        rel_triple = (u, str(rel_int), v)
        result.append(rel_triple)
    return result

edgelist = []
for u,v in subgraph.edges():
    edgelist.extend(generate_all_rel(u,v))

print("First five edges:")
print(edgelist[:5])
edgeslice = edgelist[:15]
edgearr = np.asarray(edgeslice)
test_tf = TriplesFactory.from_labeled_triples(edgearr)

#load model

rotate_model = torch.load('./rotate/best_model/trained_model.pkl')
#conve_model = torch.load('./conve/best_model/trained_model.pkl')
distmult_model = torch.load('./distmult/best_model/trained_model.pkl')
hole_model = torch.load('./hole/best_model/trained_model.pkl')

#mt_rotate = rotate_model.triples_factory.map_triples_to_id(triples=edgearr)
#mt_distmult = distmult_model.triples_factory.map_triples_to_id(triples=edgearr)
#mt_hole = hole_model.triples_factory.map_triples_to_id(triples=edgearr)

#top_15_rotate_all = get_all_prediction_df(rotate_model, k=15, triples_factory = tf)
#top_15_conve = get_all_prediction_df(conve_model, k=15, triples_factory = tf)

#top_15_distmult = get_all_prediction_df(distmult_model, k=15, triples_factory = tf)
#top_15_hole = get_all_prediction_df(hole_model, k=15, triples_factory = tf)

score_15_rotate = rotate_model.predict_scores(triples = test_tf)
score_15_distmult = distmult_model.predict_scores(triples = test_tf)
score_15_hole = hole_model.predict_scores(triples = test_tf)

print("RotatE")
print(score_15_rotate)

#print("ConvE")
#print(top_15_conve)

print("DistMult")
print(score_15_distmult)

print("Hole")
print(score_15_hole)


